import numpy as np
import pandas as pd
from ncls import NCLS  # type: ignore[import]

from pyranges.core.names import BY_ENTRY_IN_KWARGS, VALID_JOIN_TYPE


def _both_indexes(
    df: pd.DataFrame,
    df2: pd.DataFrame,
) -> tuple[pd.Index, pd.Index]:
    if df2.empty:
        return df.index, pd.Index(np.array([], dtype=np.int64))

    starts = df.Start.to_numpy()
    ends = df.End.to_numpy()
    indexes = df.index.to_numpy()

    it = NCLS(df2.Start.to_numpy(), df2.End.to_numpy(), df2.index.to_numpy())

    return it.all_overlaps_both(starts, ends, indexes)


def _both_dfs(
    df: pd.DataFrame,
    df2: pd.DataFrame,
    join_type: VALID_JOIN_TYPE,
    suffix: str,
    **kwargs,
) -> pd.DataFrame:
    columns_used_for_by = kwargs.get(BY_ENTRY_IN_KWARGS, {}).keys()

    # apply_pair has already matched df and df2, so
    #  we can drop the columns used for by (Chromosome, maybe Strand, and match_by if provided)
    df2 = df2.drop(columns=columns_used_for_by)

    _self_indexes, _other_indexes = _both_indexes(df, df2)
    expected_columns = [*df.head(0).join(df2.head(0), how="inner", rsuffix=suffix).columns]
    df2 = pd.DataFrame(df2)
    df2.columns = expected_columns[df.shape[1] :]
    _df = df.reindex(_self_indexes)
    _df.index = pd.Index(np.arange(len(_df)))
    _df2 = pd.DataFrame(df2).reindex(_other_indexes)
    _df2.index = pd.Index(np.arange(len(_df2)))
    j = _df.join(_df2, how="inner")
    if join_type == "inner":
        j.index = _self_indexes
        return j

    if join_type == "left":
        missing_rows = _missing_rows_left(_self_indexes, df, j)
        return pd.concat([j, missing_rows])

    if join_type == "right":
        missing_rows = _missing_rows_right(_other_indexes, df2, j)
        return pd.concat([j, missing_rows])

    if join_type == "outer":
        missing_rows_l = _missing_rows_left(_self_indexes, df, j)
        missing_rows_r = _missing_rows_right(_other_indexes, df2, j)
        return pd.concat([j, missing_rows_l, missing_rows_r])

    msg = f"Invalid join type: {join_type}"
    raise ValueError(msg)


def _missing_rows_left(_self_indexes, df, j) -> pd.DataFrame:
    missing_indices_left = df.index.difference(_self_indexes)
    j.index = _self_indexes
    return df.reindex(missing_indices_left)


def _missing_rows_right(_other_indexes, df2, j) -> pd.DataFrame:
    missing_indices_right = df2.index.difference(_other_indexes)
    j.index = _other_indexes
    return df2.reindex(missing_indices_right)



import pandas as pd
from sorted_nearest import annotate_clusters  # type: ignore[import]

from pyranges.core.names import END_COL, START_COL


def _cluster(
    df: pd.DataFrame,
    cluster_column: str,
    count_column: str | None = None,
    slack: int = 0,
    **_,
) -> pd.DataFrame:
    if df.empty:
        return df

    # important: sorted_nearest interprets slack differently than pyranges
    # 0 slack in sorted_nearest means that bookended intervals are clustered
    # together, while in pyranges it means that they are not.
    ids = annotate_clusters(df[START_COL].to_numpy(), df[END_COL].to_numpy(), slack=slack - 1)

    df.insert(df.shape[1], cluster_column, ids)

    if count_column:
        _count = df.groupby(cluster_column)["Cluster"].count()
        _count.name = count_column
        df = df.merge(_count, on=cluster_column)

    return df



import numpy as np
import pandas as pd

import pyranges as pr  # noqa: TCH001
from pyranges.core import names
from pyranges.core.example_data import example_data
from pyranges.core.pyranges_helpers import mypy_ensure_pyranges

Chromsizes = dict[str, int] | dict[tuple[str, str], int]


def random(
    n: int = 1000,
    length: int = 100,
    chromsizes: Chromsizes | None = None,
    seed: int | None = None,
    *,
    strand: bool = True,
) -> "pr.PyRanges":
    """Return PyRanges with random intervals.

    Parameters
    ----------
    n : int, default 1000
        Number of intervals.

    length : int, default 100
        Length of intervals.

    chromsizes : dict or DataFrame, default None, i.e. use "hg19"
        Draw intervals from within these bounds.

    strand : bool, default True
        Data should have strand.

    seed : int, default None
        Seed for random number generator.

    Examples
    --------
    >>> pr.random(seed=12345)
    index    |    Chromosome    Start      End        Strand
    int64    |    object        int64      int64      object
    -------  ---  ------------  ---------  ---------  --------
    0        |    chr4          130788360  130788460  +
    1        |    chr4          36129012   36129112   +
    2        |    chr4          69733790   69733890   -
    3        |    chr4          187723767  187723867  -
    ...      |    ...           ...        ...        ...
    996      |    chr21         13544178   13544278   -
    997      |    chr21         33556472   33556572   +
    998      |    chr21         31438477   31438577   +
    999      |    chr21         38433522   38433622   -
    PyRanges with 1000 rows, 4 columns, and 1 index columns.
    Contains 24 chromosomes and 2 strands.

    """
    rng = np.random.default_rng(seed=seed)

    df: pd.DataFrame
    if chromsizes is None:
        df = example_data.chromsizes
    elif isinstance(chromsizes, dict):
        df = pd.DataFrame({names.CHROM_COL: list(chromsizes.keys()), names.END_COL: list(chromsizes.values())})
    else:
        df = chromsizes

    p = df.End / df.End.sum()

    n_per_chrom = pd.Series(rng.choice(df.index, size=n, p=p)).value_counts(sort=False).to_frame()
    n_per_chrom.insert(1, names.CHROM_COL, df.loc[n_per_chrom.index].Chromosome)
    n_per_chrom.columns = pd.Index("Count Chromosome".split())

    random_dfs = []
    for _, (count, chrom) in n_per_chrom.iterrows():
        r = rng.integers(0, df[df.Chromosome == chrom].End - length, size=count)
        _df = pd.DataFrame({names.CHROM_COL: chrom, names.START_COL: r, "End": r + length})
        random_dfs.append(_df)

    random_df = pd.concat(random_dfs)
    if strand:
        s = rng.choice("+ -".split(), size=n)
        random_df.insert(3, "Strand", s)

    return mypy_ensure_pyranges(random_df.reset_index(drop=True))



import numpy as np
import pandas as pd
from ncls import NCLS  # type: ignore[import]


def _number_overlapping(df: pd.DataFrame, df2: pd.DataFrame, **kwargs) -> pd.DataFrame:
    keep_nonoverlapping = kwargs.get("keep_nonoverlapping", True)
    column_name = kwargs.get("overlap_col", True)

    if df.empty:
        return df
    if df2.empty:
        if keep_nonoverlapping:
            df = df.copy()
            df.insert(df.shape[1], column_name, 0)
            return df
        return df2

    oncls = NCLS(df2.Start.to_numpy(), df2.End.to_numpy(), df2.index.to_numpy())

    starts = df.Start.to_numpy()
    ends = df.End.to_numpy()
    indexes = df.index.to_numpy()

    _self_indexes, _other_indexes = oncls.all_overlaps_both(starts, ends, indexes)

    s = pd.Series(_self_indexes)
    counts_per_read = s.value_counts()[s.unique()].reset_index()
    counts_per_read.columns = ["Index", "Count"]

    df = df.copy()

    _missing_indexes = np.setdiff1d(df.index, _self_indexes)
    missing = pd.DataFrame(data={"Index": _missing_indexes, "Count": 0}, index=_missing_indexes)
    counts_per_read = pd.concat([counts_per_read, missing])

    counts_per_read = counts_per_read.set_index("Index").sort_index().squeeze()

    df.insert(df.shape[1], column_name, counts_per_read)

    if keep_nonoverlapping:
        return df
    return df[df[column_name] != 0]


def _coverage(df: pd.DataFrame, df2: pd.DataFrame, **kwargs) -> pd.DataFrame:
    fraction_col = kwargs["fraction_col"]

    if df.empty:
        return df
    if df2.empty:
        df = df.copy()
        df.insert(df.shape[1], fraction_col, 0.0)
        return df

    oncls = NCLS(df2.Start.to_numpy(), df2.End.to_numpy(), df2.index.to_numpy())

    starts = df.Start.to_numpy()
    ends = df.End.to_numpy()
    indexes = df.index.to_numpy()

    _lengths = oncls.coverage(starts, ends, indexes)
    _lengths = _lengths / (ends - starts)
    _fractions = _lengths
    _fractions = _fractions.astype("float64")
    _fractions = np.nan_to_num(_fractions)

    df = df.copy()

    df.insert(df.shape[1], fraction_col, _fractions)

    return df



import pandas as pd


def sort_one_by_one(d: pd.DataFrame, col1: str, col2: str) -> pd.DataFrame:
    """Equivalent to pd.sort_values(by=[col1, col2]), but faster."""
    d = d.sort_values(by=[col2])
    return d.sort_values(by=[col1], kind="mergesort")



from typing import TYPE_CHECKING

import pandas as pd

from pyranges.core.names import (
    BY_ENTRY_IN_KWARGS,
    END_COL,
    REVERSE_STRAND,
    START_COL,
    TEMP_CUMSUM_COL,
    TEMP_INDEX_COL,
    TEMP_LENGTH_COL,
)

if TYPE_CHECKING:
    import pyranges as pr


TOTAL_EXON_LENGTH_COL = "__temp_total_exon_len__"


def _spliced_subseq(
    df: "pr.PyRanges",
    *,
    start: int = 0,
    end: int | None = None,
    **kwargs,
) -> pd.DataFrame:
    if df.empty:
        return df

    df = df.copy()

    by_argument_given = kwargs.get("by")
    _by = kwargs.get("by", TEMP_INDEX_COL)
    by = [_by] if isinstance(_by, str) else (_by or df.index)

    strand = kwargs.get(BY_ENTRY_IN_KWARGS, {}).get("Strand")

    # at this point, strand is False if 1. spliced_subsequence was called with use_strand=False or
    #                                   2. it was called with use_strand='auto' and not self.valid_strand
    # or it can be '+' or '-' if:
    #  1. it was input as True to spliced_subsequence and passed  to pyrange_apply_single as True,
    #     which updates it to '-' or '+' before calling _spliced_subseq, or
    #  2. it was called with strand=None and self is stranded

    df.insert(df.shape[1], TEMP_LENGTH_COL, df.End - df.Start)
    df.insert(df.shape[1], TEMP_INDEX_COL, df.index)

    g = df.groupby(by, dropna=False)
    df.insert(df.shape[1], TEMP_CUMSUM_COL, g[TEMP_LENGTH_COL].cumsum())

    end = df[TEMP_CUMSUM_COL].max() if end is None else end

    minstart_idx = g[TEMP_INDEX_COL].first()

    if start < 0 or (end is not None and end < 0):
        # len_per_transc is total sum of exon length per transcript
        len_per_transc = df.loc[g[TEMP_INDEX_COL].last(), [*by, TEMP_CUMSUM_COL]].rename(
            columns={TEMP_CUMSUM_COL: TOTAL_EXON_LENGTH_COL},
        )

        # exp_len_per_transc has same rows of df with total sum of exon length
        # had to add bits to keep the order of rows right, or merge would destroy it
        if by_argument_given:
            exp_len_per_transc = (
                df.loc[:, [*by, TEMP_INDEX_COL]].merge(len_per_transc, on=by).set_index(TEMP_INDEX_COL).loc[df.index]
            )
        else:
            exp_len_per_transc = df.loc[:, by].merge(len_per_transc, on=by).set_index(TEMP_INDEX_COL).loc[df.index]  # type: ignore[reportArgumentType, reportCallIssue]

        starts = exp_len_per_transc[TOTAL_EXON_LENGTH_COL] + start if start < 0 else start

        ends = exp_len_per_transc[TOTAL_EXON_LENGTH_COL] + end if end < 0 else end
    else:
        starts = start
        ends = end

    cs_start = g[TEMP_CUMSUM_COL].shift(1, fill_value=0)
    cs_start.loc[minstart_idx] = 0

    cs_end = df[TEMP_CUMSUM_COL]

    # NOTE
    # here below, start is a scalar if originally provided > 0, or a Series if < 0
    #             end is a scalar if originally None or provided >0, or a Series if provided < 0
    if strand == REVERSE_STRAND:  # and use_strand:
        start_adjustments = starts - cs_start
        adjust_start = start_adjustments > 0
        df.loc[adjust_start, END_COL] -= start_adjustments[adjust_start].astype(df[END_COL].dtype)

        end_adjustments = cs_end - ends
        adjust_end = end_adjustments > 0
        df.loc[adjust_end, START_COL] += end_adjustments[adjust_end].astype(df[START_COL].dtype)
    else:
        start_adjustments = starts - cs_start
        adjust_start = start_adjustments > 0
        df.loc[adjust_start, START_COL] += start_adjustments[adjust_start].astype(df[START_COL].dtype)

        end_adjustments = cs_end - ends
        adjust_end = end_adjustments > 0
        df.loc[adjust_end, END_COL] -= end_adjustments[adjust_end].astype(df[END_COL].dtype)

    return df[(df[START_COL] < df[END_COL])].drop([TEMP_INDEX_COL, TEMP_LENGTH_COL, TEMP_CUMSUM_COL], axis=1)



from typing import TYPE_CHECKING

import pandas as pd

from pyranges.core.names import CHROM_COL, END_COL, START_COL, STRAND_COL

if TYPE_CHECKING:
    import pyranges as pr


def _bounds[T: ("pr.PyRanges", "pd.DataFrame")](df: T, **kwargs) -> pd.DataFrame:
    if df.empty:
        return df

    col_order = list(df.columns)

    by = kwargs.get("by")
    by = [by] if isinstance(by, str) else (by or [])

    agg_dict = agg if (agg := kwargs.get("agg")) else {}
    agg_dict.update({START_COL: "min", END_COL: "max", CHROM_COL: "first"})
    if STRAND_COL in df.columns:
        agg_dict[STRAND_COL] = "first"

    res = df.groupby(by, as_index=False).agg(agg_dict)
    return res.reindex(columns=[c for c in col_order if c in res.columns])


def _outside_bounds(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    df = df.copy()

    _chromsizes = kwargs.get("chromsizes")

    if isinstance(_chromsizes, pd.DataFrame):
        size_df = _chromsizes.df
        if not size_df.Chromosome.is_unique:
            msg = "Chromosomes must be unique in chromsizes."
            raise ValueError(msg)
        chromsizes = dict(*zip(size_df[CHROM_COL], size_df[END_COL], strict=True))
    else:
        if not isinstance(_chromsizes, dict):
            msg = "chromsizes must be a dictionary or a DataFrame"
            raise TypeError(msg)
        chromsizes = _chromsizes

    size = int(chromsizes[df.Chromosome.iloc[0]])
    clip = kwargs.get("clip", False)
    only_right = kwargs.get("only_right", False)

    ends_outright = df.End > size
    starts_outleft = df.Start < 0

    if not clip:  # i.e. remove
        df = df[~ends_outright] if only_right else df[~ends_outright & ~starts_outleft]

    else:
        starts_outright = df.Start >= size

        if only_right:
            df.loc[ends_outright, "End"] = size

            # removing intervals completely out of bounds
            df = df[~starts_outright]

        else:
            ends_outleft = df.End <= 0

            df.loc[ends_outright, "End"] = size
            df.loc[starts_outleft, "Start"] = 0

            # removing intervals completely out of bounds:
            df = df[~starts_outright & ~ends_outleft]

    return df



import glob
import pytest

def main():
    rst_files = ['docs/tutorial.rst'] + glob.glob('docs/how_to*.rst')
    # Build the command with all .rst files
    args = rst_files
    # Call pytest with the list of .rst files
    return pytest.main(args)

if __name__ == "__main__":
    import sys
    sys.exit(main())


from typing import TYPE_CHECKING

import numpy as np
from sorted_nearest import maketiles, makewindows  # type: ignore[import-untyped]

from pyranges.core.names import BY_ENTRY_IN_KWARGS, END_COL, START_COL, TEMP_END_COL, TEMP_START_COL

if TYPE_CHECKING:
    from pyranges import RangeFrame


def _windows(
    df: "RangeFrame",
    *,
    window_size: int,
    **kwargs,
) -> "RangeFrame":
    strand = kwargs.get(BY_ENTRY_IN_KWARGS, {}).get("Strand")
    # at this point, strand is False/None if 1. function was called with strand=False or
    #                                   2. it was called with strand=None and self is not stranded
    # or it can be '+' or '-' if:
    #  1. it was input as True to function and passed  to pyrange_apply_single as True,
    #     which updates it to '-' or '+' before calling _windows, or
    #  2. it was called with strand=None and self is stranded

    # for neg strand, we use the same makewindows function, so we need to reverse the coordinates
    if strand == "-":
        orig_starts = df.End.to_numpy() * -1
        orig_ends = df.Start.to_numpy() * -1
    else:
        orig_starts = df.Start.to_numpy()
        orig_ends = df.End.to_numpy()

    idxs, starts, ends = makewindows(
        df.index.to_numpy(),
        orig_starts,
        orig_ends,
        window_size,
    )

    if strand == "-":
        starts, ends = (
            ends * -1,
            starts * -1,
        )

    _df = df.reindex(idxs)

    _df.loc[:, START_COL] = starts
    _df.loc[:, END_COL] = ends

    return _df


def _tiles(
    df: "RangeFrame",
    *,
    tile_size: int,
    overlap_column: str | None,
    **_,
) -> "RangeFrame":
    if overlap_column is not None:
        df = df.copy()
        df.insert(df.shape[1], TEMP_START_COL, df.Start)
        df.insert(df.shape[1], TEMP_END_COL, df.End)

    idxs, starts, ends = maketiles(
        df.index.values,
        df.Start.values,
        df.End.values,
        tile_size,
    )

    _df = df.reindex(idxs)
    _df.loc[:, START_COL] = starts
    _df.loc[:, END_COL] = ends

    if overlap_column is not None:
        overlap = np.minimum(_df.End, _df[TEMP_END_COL]) - np.maximum(_df.Start, _df[TEMP_START_COL])
        _df.insert(_df.shape[1], overlap_column, overlap)
        _df = _df.drop_and_return([TEMP_START_COL, TEMP_END_COL], axis=1)

    return _df



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from ncls import NCLS  # type: ignore[import]

from pyranges.core.names import (
    END_COL,
    OVERLAP_ALL,
    OVERLAP_CONTAINMENT,
    OVERLAP_FIRST,
    OVERLAP_LAST,
    START_COL,
    VALID_OVERLAP_OPTIONS,
    VALID_OVERLAP_TYPE,
)
from pyranges.range_frame.range_frame import _mypy_ensure_rangeframe

if TYPE_CHECKING:
    import pyranges as pr
    from pyranges import RangeFrame


def _overlap_indices(
    df: "RangeFrame",
    df2: "RangeFrame",
    how: VALID_OVERLAP_TYPE = "first",
    **_,
) -> "pd.Series[int]":
    if df.empty or df2.empty:
        return pd.Series([], dtype=np.int64)

    starts = df.Start.to_numpy()
    ends = df.End.to_numpy()
    indexes = df.index.to_numpy()

    it = NCLS(df2.Start.to_numpy(), df2.End.to_numpy(), df2.index.to_numpy())

    if how == OVERLAP_ALL:
        _indices = it.all_overlaps_self(starts, ends, indexes)
    elif how == OVERLAP_CONTAINMENT:
        _indices, _ = it.all_containments_both(starts, ends, indexes)
    elif how == OVERLAP_FIRST:
        _indices, _ = it.first_overlap_both(starts, ends, indexes)
    elif how == OVERLAP_LAST:
        _indices, _ = it.last_overlap_both(starts, ends, indexes)
    else:
        msg = f"{VALID_OVERLAP_OPTIONS} are the only valid values for how."
        raise ValueError(msg)
    return pd.Series(_indices)


def _overlapping_indices(
    df: "pr.RangeFrame",
    df2: "pr.RangeFrame",
    how: VALID_OVERLAP_TYPE = "all",
) -> "pd.Series[int]":
    return _overlap_indices(df, df2, how)


def _overlap(
    df: "RangeFrame",
    df2: "RangeFrame",
    *,
    how: VALID_OVERLAP_TYPE = "all",
    invert: bool = False,
    **_,
) -> pd.DataFrame:
    if invert:
        df = _mypy_ensure_rangeframe(df.copy())
        df.insert(df.shape[1], "__ix__", np.arange(len(df)))

    indexes = _overlap_indices(df, df2, how)

    _result = df.reindex(indexes)

    if invert:
        found_idxs = getattr(_result, "__ix__", [])
        _result = df[~pd.Series(df.__ix__).isin(found_idxs)]
        _result = _result.drop("__ix__", axis=1, inplace=False)
    return _result


def _count_overlaps(
    df: "RangeFrame",
    df2: "RangeFrame",
    name: str,
    **_,
) -> pd.DataFrame:
    idx = _overlapping_indices(df, df2)
    vc: "pd.Series[int]" = idx.value_counts(sort=False)
    sx = pd.DataFrame(np.zeros(len(df), dtype=np.int64), index=df.index)

    sx.loc[vc.index, 0] = vc.to_numpy()

    df.insert(df.shape[1], name, sx.squeeze())
    return df


def _intersect(
    df: "RangeFrame",
    df2: "RangeFrame",
    *,
    how: VALID_OVERLAP_TYPE = "all",
    **_,
) -> pd.DataFrame:
    df = df.copy()
    starts = df.Start.to_numpy()
    ends = df.End.to_numpy()
    indexes = df.index.to_numpy()

    in_dtype = df2.Start.dtype

    oncls = NCLS(df2.Start.to_numpy(), df2.End.to_numpy(), df2.index.to_numpy())

    if how == OVERLAP_ALL:
        _self_indexes, _other_indexes = oncls.all_overlaps_both(starts, ends, indexes)
    elif how == OVERLAP_CONTAINMENT:
        _self_indexes, _other_indexes = oncls.all_containments_both(starts, ends, indexes)
    elif how == OVERLAP_FIRST:
        _self_indexes, _other_indexes = oncls.first_overlap_both(starts, ends, indexes)
    elif how == OVERLAP_LAST:
        _self_indexes, _other_indexes = oncls.last_overlap_both(starts, ends, indexes)
    else:
        msg = "Invalid overlap type. Valid types are: first, containment, all, last."
        raise ValueError(msg)

    df, df2 = df.reindex(_self_indexes), df2.reindex(_other_indexes)

    new_starts = pd.Series(
        np.where(df.Start.to_numpy() > df2.Start.to_numpy(), df.Start, df2.Start),
        index=df.index,
        dtype=in_dtype,
    )

    new_ends = pd.Series(
        np.where(df.End.to_numpy() < df2.End.to_numpy(), df.End, df2.End),
        index=df.index,
        dtype=in_dtype,
    )

    df.loc[:, START_COL] = new_starts
    df.loc[:, END_COL] = new_ends

    return df



import pandas as pd
from sorted_nearest import find_clusters  # type: ignore[import]

from pyranges.core.names import END_COL, START_COL


def _merge(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    if df.empty:
        return df

    slack = kwargs.get("slack", 0)
    by = kwargs["by"]

    cdf = df.sort_values(START_COL)

    # important: sorted_nearest interprets slack differently than pyranges
    # 0 slack in sorted_nearest means that bookended intervals are considered overlapping
    # together, while in pyranges it means that they are not.
    starts, ends, number = find_clusters(cdf.Start.values, cdf.End.values, slack - 1)

    by_values = df.head(1).squeeze()[by].to_dict()

    cluster_df = pd.DataFrame(
        {
            START_COL: starts,
            END_COL: ends,
        }
        | by_values,
    )
    # Sort columns in the original order of the dataframe.
    cluster_df = cluster_df[[c for c in cdf.columns if c in cluster_df]]

    if kwargs["count_col"]:
        cluster_df.insert(cluster_df.shape[1], kwargs["count_col"], number)

    return cluster_df



import math
from collections.abc import Generator

import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame

from pyranges.core.names import BY_ENTRY_IN_KWARGS, END_COL, FORWARD_STRAND, REVERSE_STRAND, START_COL


def run_in_parallel(function, dfs: list[DataFrame], nb_cpu: int, *args, **kwargs) -> Generator:
    """Run a function in parallel on a list of DataFrames."""
    # TODO(endbak): Things to consider:  # noqa: FIX002
    #   https://github.com/endrebak/pyranges1_alpha/issues/21
    #   The index will be different here than in the original DataFrame if the function resets the index
    #   Perhaps we should not concat so the function can return anything?
    from joblib import Parallel, delayed

    return Parallel(n_jobs=nb_cpu)(delayed(function)(df, *args, **kwargs) for df in dfs)  # type: ignore[return-type]


def split_df_into_chunks_without_splitting_groups(
    df: DataFrame,
    *,
    by: list[str],
    nb_splits: int,
) -> list[DataFrame]:
    """Split a DataFrame into chunks without splitting groups.

    Args:
    ----
        df: DataFrame to split.
        by: Column(s) to group by.
        nb_splits: Number of splits.

    Returns:
    -------
        List of DataFrames.

    """
    # Calculate the target size for each chunk
    target_chunk_size = math.ceil(df.shape[0] / nb_splits)

    # Assign a group ID and calculate cumsum of rows per group
    df = df.sort_values(by)
    groupby = df.groupby(by)

    chunk_ids = groupby.size().cumsum().floordiv(target_chunk_size)
    group_sizes = groupby.size()
    df = df.assign(__chunk_id__=np.repeat(a=chunk_ids.to_numpy(), repeats=group_sizes.to_numpy()))

    # Split the DataFrame using groupby
    return [group.drop("__chunk_id__", axis="columns") for _, group in df.groupby("__chunk_id__")]


def _lengths(df: DataFrame) -> pd.Series:
    return df[END_COL] - df[START_COL]


def _tss(df: DataFrame, **kwargs) -> DataFrame:
    df = df.copy(deep=True)
    dtype = df.dtypes["Start"]
    slack = kwargs.get("slack", 0)

    starts = np.where(df.Strand == "+", df.Start, df.End - 1)
    ends = starts + slack + 1
    starts = starts - slack
    starts = np.where(starts < 0, 0, starts)

    df.loc[:, "Start"] = starts.astype(dtype)
    df.loc[:, "End"] = ends.astype(dtype)

    return df


def _tes(df: DataFrame, **kwargs) -> DataFrame:
    df = df.copy(deep=True)
    dtype = df.dtypes["Start"]
    slack = kwargs.get("slack", 0)

    starts = np.where(df.Strand == "+", df.End - 1, df.Start)
    ends = starts + 1 + slack
    starts = starts - slack
    starts = np.where(starts < 0, 0, starts)

    df.loc[:, "Start"] = starts.astype(dtype)
    df.loc[:, "End"] = ends.astype(dtype)

    return df


def _extend(
    df: DataFrame,
    ext: int | None = None,
    ext_3: int | None = None,
    ext_5: int | None = None,
    **kwargs,
) -> DataFrame:
    df = df.copy()
    dtype = df.Start.dtype

    if ext is not None:
        df.loc[:, START_COL] = df.Start - ext
        df.loc[df.Start < 0, START_COL] = 0
        df.End = df.End + ext
    else:
        strand = kwargs.get(BY_ENTRY_IN_KWARGS, {}).get("Strand")
        strand = strand if strand else FORWARD_STRAND

        if ext_5 and strand == FORWARD_STRAND:
            df.loc[:, START_COL] -= ext_5
        elif ext_5 and strand == REVERSE_STRAND:
            df.loc[:, END_COL] += ext_5

        if ext_3 and strand == REVERSE_STRAND:
            df.loc[:, START_COL] -= ext_3
        elif ext_3 and strand == FORWARD_STRAND:
            df.loc[:, END_COL] += ext_3

    df = df.astype({START_COL: dtype, END_COL: dtype})

    if not (df.Start < df.End).all():
        msg = "Some intervals are negative or zero length after applying extend!"
        raise ValueError(msg)

    return df


def _extend_grp(
    df: DataFrame,
    group_by: list[str] | str,
    ext: int | None = None,
    ext_3: int | None = None,
    ext_5: int | None = None,
    **kwargs,
) -> DataFrame:
    df = df.copy()
    g = df.groupby(group_by)

    minstarts_pos = g.Start.idxmin()
    maxends_pos = g.End.idxmax()

    if ext is not None:
        df.loc[minstarts_pos, START_COL] = df.Start - ext
        df.loc[df.Start < 0, START_COL] = 0
        df.loc[maxends_pos, END_COL] = df.End + ext

    else:
        strand = kwargs.get(BY_ENTRY_IN_KWARGS, {}).get("Strand")
        strand = strand if strand else FORWARD_STRAND

        if ext_5 and strand == FORWARD_STRAND:
            df.loc[minstarts_pos, START_COL] -= ext_5
        elif ext_5 and strand == REVERSE_STRAND:
            df.loc[maxends_pos, END_COL] += ext_5

        if ext_3 and strand == REVERSE_STRAND:
            df.loc[minstarts_pos, START_COL] -= ext_3
        elif ext_3 and strand == FORWARD_STRAND:
            df.loc[maxends_pos, END_COL] += ext_3

    if not (df.Start < df.End).all():
        msg = "Some intervals are negative or zero length after applying extend!"
        raise ValueError(msg)

    return df
